@@ -14,11 +14,11 @@
 
 import logging
 
-DEFAULT_API_URL = "https://arweave.net"
+DEFAULT_API_URL = 'https://arweave.net'
 
 logger = logging.getLogger(__name__)
 
-class ArweaveTransactionException(Exception):
+class ArweaveException(Exception):
     pass
 
 from .peer import Peer
