@@ -19,7 +19,7 @@ from . import logger
 class Transaction(object):
     def __init__(self, wallet, **kwargs):
         self.jwk_data = wallet.jwk_data
-        self.jwk = jwk.construct(self.jwk_data, algorithm="RS256")
+        self.jwk = jwk.construct(self.jwk_data, algorithm='RS256')
         self.wallet = wallet
 
         self.id = kwargs.get('id', '')
@@ -52,7 +52,7 @@ class Transaction(object):
         if kwargs.get('transaction'):
             self.from_serialized_transaction(kwargs.get('transaction'))
         else:
-            self.data_root = ""
+            self.data_root = ''
 
             self.data_tree = []
 
@@ -65,8 +65,8 @@ class Transaction(object):
             self.quantity = kwargs.get('quantity', '0')
             if float(self.quantity) > 0:
                 if self.target == '':
-                    raise ArweaveTransactionException(
-                        "Unable to send {} AR without specifying a target address".format(self.quantity))
+                    raise ArweaveException(
+                        'Unable to send {} AR without specifying a target address'.format(self.quantity))
 
                 # convert to winston
                 self.quantity = ar_to_winston(float(self.quantity))
@@ -86,10 +86,12 @@ class Transaction(object):
 
     def from_serialized_transaction(self, transaction_json):
         if type(transaction_json) == str:
-            self.load(json.loads(transaction_json))
+            transaction_json = json.loads(transaction_json)
+        if type(transaction_json) == dict:
+            self.load(transaction_json)
         else:
-            raise ArweaveTransactionException(
-                "Please supply a string containing json to initialize a serialized transaction")
+            raise ArweaveException(
+                'Please supply a string or dict containing json to initialize a serialized transaction')
 
     def get_reward(self, data_size, target_address=None):
         reward = self.peer.price(data_size, target_address)
@@ -121,7 +123,7 @@ class Transaction(object):
     def get_signature_data(self):
         self.reward = self.get_reward(self.data_size, target_address=self.target if len(self.target) > 0 else None)
 
-        if int(self.data_size) > 0 and self.data_root == "" and not self.uses_uploader:
+        if int(self.data_size) > 0 and self.data_root == '' and not self.uses_uploader:
             if type(self.data) == str:
                 root_hash = compute_root_hash(io.BytesIO(base64url_decode(self.data.encode('utf-8'))))
 
@@ -131,11 +133,11 @@ class Transaction(object):
             self.data_root = base64url_encode(root_hash)
 
         if self.format == 1:
-            tag_str = ""
+            tag_str = ''
 
             for tag in self.tags:
                 name, value = decode_tag(tag)
-                tag_str += "{}{}".format(name.decode(), value.decode())
+                tag_str += '{}{}'.format(name.decode(), value.decode())
 
             owner = base64url_decode(self.jwk_data['n'].encode())
             target = base64url_decode(self.target)
@@ -153,7 +155,7 @@ class Transaction(object):
             tag_list = [[tag['name'].encode(), tag['value'].encode()] for tag in self.tags]
 
             signature_data_list = [
-                "2".encode(),
+                '2'.encode(),
                 base64url_decode(self.jwk_data['n'].encode()),
                 base64url_decode(self.target.encode()),
                 str(self.quantity).encode(),
@@ -170,7 +172,7 @@ class Transaction(object):
     def send(self):
         try:
             self.peer.send_tx(self.json_data)
-        except ArweaveTransactionException:
+        except ArweaveException:
             pass
 
         return self.last_tx
@@ -199,7 +201,7 @@ class Transaction(object):
             if len(self.data_root) > 0:
                 data['data_root'] = self.data_root.decode()
             else:
-                data['data_root'] = ""
+                data['data_root'] = ''
             data['data_size'] = str(self.data_size)
             data['data_tree'] = []
 
@@ -218,21 +220,21 @@ class Transaction(object):
     def get_status(self):
         try:
             self.status = self.peer.tx_status(self.id)
-        except ArweaveTransactionException:
-            self.status = "PENDING"
+        except ArweaveException:
+            self.status = 'PENDING'
         return self.status
 
     def get_transaction(self):
         try:
             self.load(self.peer.tx(self.id))
-        except ArweaveTransactionException as exception:
+        except ArweaveException as exception:
             pass
 
     def get_price(self):
         try:
             price = self.peer.price(self.data_size)
             return winston_to_ar(price)
-        except TransactionException as exception:
+        except ArweaveException as exception:
             pass
 
     def get_data(self):
@@ -261,16 +263,16 @@ class Transaction(object):
 
         if not self.chunks:
             self.chunks = {
-                "chunks": [],
-                "data_root": b'',
-                "proof": []
+                'chunks': [],
+                'data_root': b'',
+                'proof': []
             }
 
             self.data_root = ''
 
     def get_chunk(self, idx):
         if self.chunks is None:
-            raise ArweaveTransactionException("Chunks have not been prepared")
+            raise ArweaveException('Chunks have not been prepared')
 
         proof = self.chunks.get('proofs')[idx]
         chunk = self.chunks.get('chunks')[idx]
@@ -280,9 +282,9 @@ class Transaction(object):
         chunk_data = self.file_handler.read(chunk.data_size)
 
         return {
-            "data_root": self.data_root.decode(),
-            "data_size": str(self.data_size),
-            "data_path": base64url_encode(proof.proof),
-            "offset": str(proof.offset),
-            "chunk": base64url_encode(chunk_data)
+            'data_root': self.data_root.decode(),
+            'data_size': str(self.data_size),
+            'data_path': base64url_encode(proof.proof),
+            'offset': str(proof.offset),
+            'chunk': base64url_encode(chunk_data)
         }
